#include <stdio.h>
#include <stdlib.h>

struct cell {
    int           data;
    struct cell * next;
    struct cell * prev; 
};
typedef struct cell cell_t;
typedef struct {
    cell_t * start; // pointeur sur le premier lment 
    cell_t * end;   // pointeur sur le dernier lment 
} list_t;

    /* Cration d'une liste vide. */
list_t * create_list() {
    list_t * L = (list_t *) malloc(sizeof(list_t));
    if (L != NULL) {
	L->start = NULL;
	L->end   = NULL;
    }
    return L;
}
    /* Affiche le contenu de la liste L. */
void print_list(list_t * L) {
    if (L == NULL) return;
    cell_t * tmp = L->start;
    while(tmp != NULL) { 
        printf("%2d -> ", tmp->data);
        tmp = tmp->next;
    }
    printf("NULL\n");
}
    /* Renvoie le nombre d'lments de la liste. */
long get_size(list_t * L) {
    if (L == NULL) return -1;
    cell_t * tmp = L->start;
    long res = 0; 
    for(; tmp != NULL; tmp=tmp->next,res++);
    return res;
}
    /* Insre un lment en tete de liste. */
int insert_in_head(int element, list_t * L) {
    if (L == NULL) return -2;
    // Cration de la nouvelle cellule.
    cell_t * new_cell = (cell_t *) malloc(sizeof(cell_t));
    if (new_cell == NULL) return -1;
    new_cell->data = element;
    new_cell->next = L->start; // Insertion en tte.
    new_cell->prev = NULL;
    // Deuxime tape: mise--jour de la liste et des cellules voisines.
    if (L->start != NULL) (L->start)->prev = new_cell;
    if (L->end == NULL) L->end = new_cell;  
    L->start = new_cell;
    return 0;
}
    /* Insre un lment en queue de liste. */
int insert_in_tail(int element, list_t * L) {
    if (L == NULL) return -2;
    // Cration de la nouvelle cellule.
    cell_t * new_cell = (cell_t *) malloc(sizeof(cell_t));
    if (new_cell == NULL) return -1;
    new_cell->data = element;
    new_cell->next = NULL; // Insertion en tte.
    new_cell->prev = L->end;
    // Deuxime tape: mise--jour de la liste et des cellules voisines.
    if (L->end   != NULL) (L->end)->next = new_cell;
    if (L->start == NULL) L->start = new_cell;  
    L->end = new_cell;
    return 0;
}
    /* Suppression du contenu de la liste L. */
void free_content_of(list_t * L) {
    if (L == NULL) return; 
    cell_t * tmp = L->start, *cur = NULL;
    while ((cur = tmp) != NULL) { // cur: pointeur sur la cellule courante.
	tmp = tmp->next;
	free(cur);  // Libration des lments.
    }
    L->start = NULL;
    L->end   = NULL;
}
    /* Suppression de la liste L et de son contenu. */
void destroy(list_t * L) {
    free_content_of(L);
    free(L);
}
    /* Ajout au bon endroit de e dans la liste ordonne;
       renvoie 0 en cas de succs, 1 en cas d'erreur (e dj prsent). */
int insert(int e, list_t * L) {
    if (L == NULL) return -2;
    if ((L->start == NULL) || ((L->start)->data > e)) {
	// Cas liste vide ou premier lment plus grand que e.
	if (! insert_in_head(e, L)) return 0;
	else return -1;
    }
    if ((L->end)->data < e) {
	if (! insert_in_tail(e, L)) return 0;
	else return -1;
    }
    // Ici, e est dans l'intervalle des valeurs de la liste.
    cell_t * tmp = L->start;
    while(tmp->data < e) tmp = tmp->next; // Dplacement jusqu'au bon endroit.
    if (tmp->data == e) return -1;
    // Cration de la nouvelle cellule
    cell_t * new_cell = (cell_t *) malloc(sizeof(cell_t));
    if (new_cell == NULL) return -1;
    new_cell->data = e;
    new_cell->next = tmp; 
    new_cell->prev = tmp->prev;
    // Deuxime tape: mise--jour des cellules voisines.
    (tmp->prev)->next = new_cell;
    tmp->prev = new_cell;
    return 0;
}
    /* Suppression de l'entier e de la liste ordonne;
       renvoie 0 en cas de succs, 1 en cas d'erreur (e absent de la liste). */
int delete(int e, list_t * L) {
    if (L == NULL) return -2;
    if ((L->start == NULL) || (L->start->data > e) || (L->end->data < e))
	return -1;
    cell_t * tmp = L->start;
    //Cas "suppression du premier lment": il faut au moins mettre  jour L->start.
    if (tmp->data == e) {
	    L->start = tmp->next;
	    if (L->start != NULL) L->start->prev = NULL; 
	    else L->end = NULL; // On avait une liste  un lment.
	    free(tmp);
	    return 0;
    }
    // Cas "suppression du dernier lment": il faut au moins mettre  jour L->end.
    tmp = L->end;
    if (tmp->data == e) {
	    L->end = tmp->prev;
	    if (L->end != NULL) L->end->next = NULL;
	    else L->start = NULL; // On avait une liste  un lment
	    free(tmp);
	    return 0;
    }
    // Cas "suppression d'un lment intermdiaire".
    tmp = L->start->next;
    while(tmp->data < e) tmp = tmp->next;     // On va au bon endroit.
    if (tmp->data != e) return -1;  // e n'est pas dans la liste.
    cell_t * tmp_prev = tmp->prev;
    cell_t * tmp_next = tmp->next;
    tmp_prev->next = tmp_next;
    tmp_next->prev = tmp_prev;
    free(tmp);
    return 0;
}
     /* Fonction qui facilite les tests d'insertion. */
void test_add(int tab[], size_t n, list_t * L) {
    size_t i;
    for (i=0; i<n; i++) {
	printf("Ajout de %2d : [%d]",tab[i],insert(tab[i],L));
	printf(" --> |L|=%lu et L = ", get_size(L));  
	print_list(L);
    }
}
     /* Fontion qui facilite les tests de suppression. */
void test_delete(int tab[], size_t n, list_t * L) {
    size_t i;
    for (i=0; i<n; i++) {
	printf("Suppression de %2d : [%d]",tab[i],delete(tab[i],L));
	printf(" --> |L|=%lu et L = ", get_size(L));  
	print_list(L);
    }
}

int main() 
{
    list_t * L = create_list();
    int to_add[] = { 13, 10, 14, 11, 9, 12,     
		   13, 12, 11 }; // testent l'ajout de valeurs dj prsentes
    int to_del[] = { 9, 14, 11, 5, 15, 11};
    printf("*** tests d'insertion (listes non ordonnes) ***\n");
    insert_in_head(10, L);
    insert_in_head(9,  L);
    insert_in_tail(10, L);
    insert_in_tail(9,  L);
    printf(" -> L = "); print_list(L);
    free_content_of(L);
    printf("\n***   tests d'insertion (listes ordonnes)  ***\n");    
    test_add(to_add, sizeof(to_add)/sizeof(int), L);
    printf("\n*** tests de suppression (listes ordonnes) ***\n");    
    test_delete(to_del, sizeof(to_del)/sizeof(int), L);
    destroy(L);
    return 0;
}
