#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

    /* Dfinition de la structure cell */
struct cell {
    int           data;
    struct cell * next;
};
typedef struct cell * list_t; // le type list_t permet de reprsenter une liste
typedef struct cell   cell_t; // le type cell_t permet de reprsenter une cellule

    /* Insre un lment en tte de liste - renvoie la liste complte */
list_t insert_in_head(int element, list_t L) {
    // Cration de la nouvelle cellule
    cell_t * new_cell = (cell_t *) malloc(sizeof(cell_t));
    if (new_cell == NULL) return NULL;
    new_cell->data = element;
    new_cell->next = L; // insertion en tte
    // La liste rsultat est en fait la premire cellule
    return (list_t) new_cell; // le cast est facultatif
}

    /* insertion en tte - version procdurale */
void proc_insert_in_head(int element, list_t * L) {
    list_t* old = L;
    *L = insert_in_head(element, *L);
    if (L == NULL) *L = *old; // au cas ou l'insertion a chou
}

    /* Insre un lment en queue de liste - renvoie la liste complte */
list_t insert_in_tail(int element, list_t L) {
    // Cration de la nouvelle cellule
    cell_t * new_cell = (cell_t *) malloc(sizeof(cell_t));
    if (new_cell == NULL) return NULL;
    new_cell->data = element;
    new_cell->next = NULL; // insertion en queue
    // recherche de l'endroit o effectuer l'insertion
    if (L == NULL) return (list_t) new_cell;
        //  partir de l, L!=NULL et L->next a un sens
    list_t tmp = L; 
    while (tmp->next != NULL) tmp = tmp->next; // on va  la dernire cellule
    // on effectue l'insertion et on renvoie la liste complte
    tmp->next = (list_t) new_cell; // le cast est facultatif
    return L;
}

    /* insertion en queue - version procdurale */
void proc_insert_in_tail(int element, list_t * L) {
    list_t* old = L;
    *L = insert_in_tail(element, *L);
    if (L == NULL) *L = *old;
}

    /* suppression de l'lment en tte de liste */
list_t delete_in_head(list_t L) {
    if (L == NULL) return L;
    list_t res = L->next;
    free(L); // libration de la premire cellule   
    return res;  
} 

    /* suppression de l'lment en tte de liste - version procdurale */
void proc_delete_in_head(list_t * L) {
    *L = delete_in_head(*L);
} 

    /* suppression de l'lment en queue de liste */
list_t delete_in_tail(list_t L) {
    if (L == NULL) return L;
    if (L->next == NULL) { // liste  un seul lment
        free(L);
        return NULL;
    }
    list_t tmp    = L->next;
    cell_t * prec = L; // pointeur sur la cellule prcdant tmp
    while(tmp->next != NULL) {
        prec = tmp;
        tmp  = tmp->next;
    }  
    // ici, tmp est la dernire cellule (en particulier, prec->next est tmp)
    prec->next = NULL; // mise  jour de la cellule voisine
    free(tmp);         // libration de la dernire cellule   
    return L;  
} 

    /* suppression de l'lment en queue de liste - version procdurale */
void proc_delete_in_tail(list_t * L) {
    *L = delete_in_tail(*L); // pour changer :-)
}

    /* suppression de la liste complete */
void delete_list(list_t *L) {
    while(*L != NULL) proc_delete_in_head(L);
}

    /* affiche le contenu de la liste L */
void print_list(list_t L) {
    while(L != NULL) { // on aurait pu crire 'while (L)'
        printf("%d -> ", L->data);
        L = L->next;
    }
    printf("NULL\n");
}

list_t insert_in_head     (int element, list_t L);
list_t insert_in_tail     (int element, list_t L);
void   proc_insert_in_head(int element, list_t * L);
void   proc_insert_in_tail(int element, list_t * L);

list_t delete_in_head     (list_t L);
list_t delete_in_tail     (list_t L);
void   proc_delete_in_head(list_t * L);
void   proc_delete_in_tail(list_t * L);

void delete_list(list_t *L);

int main()
{
    /* Cration de la liste L = (10, 11, 12, 13, 14, 15) */
    list_t L = insert_in_head(11, insert_in_head(12, NULL)); // avec insert_in_head
    proc_insert_in_head(10, &L);                             // version procdurale
    L = insert_in_tail(14,  insert_in_tail(13, L)); // ajout en queue de liste
    proc_insert_in_tail(15, &L);                    // version procdurale
    printf("L = "); print_list(L);
    
    /* Suppression d'lment en tte */
    L = delete_in_head(L);   
    proc_delete_in_head(&L);                        // version procdurale
    printf("Suppression de deux lments en tte  : L = ");
    print_list(L);

    /* Suppression d'lment en queue */
    L = delete_in_tail(L);   
    proc_delete_in_tail(&L);                        // version procdurale
    printf("Suppression de deux lments en queue : L = ");
    print_list(L);
    
    /* Supression de la liste */
    delete_list(&L);
    printf("Suppression de la liste : L = ");
    print_list(L);
    return 0;
}




