#include <stdio.h>
#include <string.h>

struct numero_SS_t {
    unsigned char sexe; 
    unsigned char annee;
    unsigned char mois; 
    unsigned char dept;
    unsigned int  commune;
    unsigned int  ordre;
    unsigned char clef;
};

void print_mois(char m) {
    const char * tab[] = {
	"janvier", "fvrier", "mars", "avril", "mai", "juin", "juillet", 
	"aot", "septembre", "octobre", "novembre", "dcembre"
    };
    if ((m < 1) || (m > 12)) printf("INCONNU");
    else printf("%s", tab[m]);
}

int str2int(char * s, unsigned inf, unsigned sup) {
    if ((inf > sup) || (sup >= strlen(s))) return -1;
    int res = 0, i;
    for (i=inf; i<= sup; i++) res = res*10 + (s[i] - '0');
    return res;
}

int compute_key(struct numero_SS_t * nss) {
    unsigned long r1 = (nss->dept + (nss->mois * 100) + (nss->annee * 10000) + 
			(nss->sexe * 1000000)) % 97;
    unsigned long res = 
	97 - ((r1 * 1000000) + (nss->commune * 1000) + nss->ordre) % 97;
    return res;	
}

int check_bad_NSS_format(char * s) {
    if ((s == NULL) || (strlen(s) != 15)) return -1;
    int i;
    for (i=0; i<15; i++) 
	if ((s[i] < '0') || s[i] > '9') return -1; 
    return 0;
}

int getNSSfrom(char * s, struct numero_SS_t * nss) {
    if ((nss == NULL) || check_bad_NSS_format(s)) return -1;
    nss->sexe    = s[0] - '0';
    nss->annee   = str2int(s,1,2);
    nss->mois    = str2int(s,3,4);
    nss->dept    = str2int(s,5,6);
    nss->commune = str2int(s,7,9);
    nss->ordre   = str2int(s,10,12);
    nss->clef    = str2int(s,13,14);
    return 0;
}

void getInfoFrom(struct numero_SS_t * nss) {
    if (compute_key(nss) == nss->clef) {
	printf("%s en ",(nss->sexe == 1)?"Homme n":"Femme ne");
	print_mois(nss->mois);
	printf(" %d dans le dpartement %d\n", 
	       nss->annee, nss->dept);
    } else printf("Clef de contrle incorrecte\n");
}

void printNSS(struct numero_SS_t * nss) {
    printf("%d %d %d %d %d %d -- %d\n", nss->sexe, nss->annee,
	   nss->mois,nss->dept ,nss->commune,nss->ordre,nss->clef);
}

int main()
{
    struct numero_SS_t nss; // = {2, 53, 7, 75, 73, 4, 83};
    char * s = "253077507300483";
    if (getNSSfrom(s,&nss) == -1) return 1;
    printNSS(&nss);
    getInfoFrom(&nss);
    return 0;
}
