#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>

#define A_SIZE 26  // taille de l'alphabet considr

typedef struct {
    char   c;    // le caractre
    double freq; // la frquence d'apparition, en %
} letter_freq_t ;

letter_freq_t TAB_REF_FR[A_SIZE] = { // statistique des frquences (texte franais)
    {'a', 8.11}, {'b', 0.81}, {'c', 3.38}, {'d', 4.28}, {'e', 17.69}, {'f', 1.13}, 
    {'g', 1.19}, {'h', 0.74}, {'i', 7.24}, {'j', 0.18}, {'k', 0.02},  {'l', 5.99},
    {'m', 2.29}, {'n', 7.68}, {'o', 5.20}, {'p', 2.92}, {'q', 0.83},  {'r', 6.43},
    {'s', 8.87}, {'t', 7.44}, {'u', 5.23}, {'v', 1.28}, {'w', 0.06},  {'x', 0.53},
    {'y', 0.26}, {'z', 0.12 }
};

size_t count_letters_in(char * file, int tab[A_SIZE]) {
    assert(A_SIZE >= 26); // cf /*@\S\ref{std::assert}@*/
    FILE * fi = fopen(file, "r");
    if ( fi == NULL ) {
	perror("[Erreur] fopen");
	return 1; 
    }
    int c;
    size_t res = 0;
    while (1) {
	c = getc(fi); // on ne teste pas les erreurs de lecture
	if ( feof(fi) ) break;
	c = tolower(c); // on se ramne  des lettres minuscules
	if ( (c >= 'a') && (c <= 'z') ) {
	    tab[c - 'a']++;
	    res++;
	}
    }
    fclose(fi);
    return res; 
}

void analyse_freq(char * file, letter_freq_t tab[A_SIZE]) {
    int * tmp_tab = (int *) calloc(A_SIZE, sizeof(int));
    size_t i, nb_char = count_letters_in(file, tmp_tab);
    for (i=0; i<A_SIZE; i++) {
	tab[i].c    = i + 'a';
	tab[i].freq = ((double) tmp_tab[i]*100)/nb_char;
    }
    free(tmp_tab);
}

int compare(const void * a, const void * b) { // pour qsort
    double freq_a = ((letter_freq_t *) a)->freq, 
	freq_b = ((letter_freq_t *) b)->freq;
    return (freq_a < freq_b) ? 1 : (freq_a > freq_b) ? -1 : 0;
}

void sort_by_letter_freq(letter_freq_t tab[A_SIZE]) { // tri (frq. dcroissantes)
    qsort(tab, A_SIZE, sizeof(letter_freq_t), compare);
}

// Affichage pour comparaison entre les frquences obtenues et la rfrence
void show_freq_comparison(letter_freq_t tab[A_SIZE], int nb_line_to_show) {
    letter_freq_t * tab_ref = TAB_REF_FR;
    int i, limit = (nb_line_to_show < A_SIZE)?nb_line_to_show:A_SIZE;
    sort_by_letter_freq(tab_ref);
    printf("---------------+----------------\n");
    printf(" Texte analys | Rfrence (fr)\n");
    printf("---------------+----------------\n");
    for (i=0; i<limit; i++) {
	printf("  %c :  %5.2f%%  |", tab[i].c,     tab[i].freq);
	printf("  %c :  %2.2f%%\n",  tab_ref[i].c, tab_ref[i].freq);
    }
    if (limit < A_SIZE) printf("             [...]\n");
}

int main(int argc, char * argv[])
{
    if (argc != 2) {
	printf("usage: %s filename\n",argv[0]);
	return EXIT_FAILURE;
    }
    letter_freq_t tab_freq[A_SIZE];
    analyse_freq(argv[1], tab_freq);    // analyse frquentielle
    sort_by_letter_freq(tab_freq);      // tri par frquence
    show_freq_comparison(tab_freq, 10); // affichage pour comparaison
    return 0;
}
