/**********************************************************
 * @file tools.c
 * @author Sebastien Varrette <Sebastien.Varrette@imag.fr>
 * @date   Mon Apr 10 2006  
 *
 * Copyright (c) 2006 Sebastien Varrette  (http://www-id.imag.fr/~svarrett/)
 *
 * @brief  
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Sebastien Varrette                               \n
 * <Sebastien.Varrette@imag.fr>                     \n
 * University of Luxembourg / ID-IMAG Laboratory    \n
 * 162-A avenue de la faiencerie                    \n
 * L-1511 Luxembourg, LUXEMBOURG                    \n
 */
 /********************************************************************************/

#include <stdio.h>

/**
 * Calcul du pgcd de 2 nombres par l'algorithme d'Euclide
 */
int pgcd(int a, int b) {
    int e = (a < 0)?-a:a; // Pour assurer des valeurs positives
    int f = (b < 0)?-b:b;
    int tmp;
    if (e == 0) return f;
    while (f > 0) {
	tmp   = f;
	f = e % f; // Calcul de r = a % b
	e = tmp;
    }
    return e;
}

/**
 * Version rcursive du calcul du pgcd
 */
int pgcd_rec(int a, int b) {
    int e = (a < 0)?-a:a;
    int f = (b < 0)?-b:b;
    if (f == 0) return e;
    return pgcd_rec(f, e%f);
}

/**
 * Calcul des coefficients de Bezout tels que d = pgcd(a,b) = a.u + b.v
 * @return   d = pgcd(a,b)
 */
int bezout(int a, int b, int * u, int * v) {
    int e = (a < 0)?-a:a;
    int f = (b < 0)?-b:b;
    *u = 1;
    *v = 0;
    int u_aux = 0, v_aux = 1;
    int q, r, tmp;
    while (f > 0) {
	q = e/f;
	r = e%f;
	e = f;
	f = r; 
	tmp = u_aux; u_aux = *u - q*u_aux; *u = tmp;
	tmp = v_aux; v_aux = *v - q*v_aux; *v = tmp;
    }
    if (a < 0) *u = -(*u);
    if (b < 0) *v = -(*v);
    return e;
}
/**
 * Calcul de l'inverse de a modulo n.
 * @return   0 en cas de succs, 1 en cas d'echec (si pgcd(a,n)=1)
 */
int inv_mod(int a, int n, int * res) {
    int d, u, v;
    d = bezout(a, n, &u, &v);
    if (d == 1) {
	*res = (n + u) % n;
	return 0;
    }
    return 1;
}
