#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

inline void error(char * msg) { // Affichage d'un message d'erreur et sortie.
        fprintf(stderr, "[Erreur] %s: %s\n", msg, strerror(errno));
        exit(EXIT_FAILURE);
}

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/uio.h>

#define STR_SIZE 20             // Taille maximale d'une chaine de caractres.
#define MAX_PERS 100            // Taille maximale de l'annuaire.
#define ANNUAIRE "annuaire.txt" // Le nom du fichier qui contient l'annuaire.

struct personne {
	char nom[STR_SIZE], prenom[STR_SIZE], tel[STR_SIZE];
	int8_t age;
	char align[3]; /* Pour l'alignement. */
};

int main()
{
	int err = 0;
	struct personne tab[MAX_PERS] = {
		{"Varrette", "Sbastien", "+33 (0)0 11 22 33 44", 26},
		{"Bernard",  "Nicolas",   "+33 (0)0 55 66 77 88", 25},
		{"", "", "", -1} // Entre spciale (indique la fin de l'annuaire).
	};

	/* Sauvegarde de l'annuaire. */
	int fd = open(ANNUAIRE, O_CREAT | O_RDWR, 0600);
	if (fd == -1) error("open (write)");
	uint16_t i = 0;
	while (strlen(tab[i].nom) > 0) i++; // Calcul du nombre d'entres.
	err = write(fd, &i, 2); // criture du nombre d'entres au dbut du fichier.
	if (err != 2) error("write (1)");

        /* On s'assure que le fichier est assez grand avant d'utiliser mmap. */
	err = lseek(fd, i * sizeof(struct personne) + 2, SEEK_SET);
	if (err == -1) error("lseek");
	err = write(fd, &i, 2);
	if (err != 2) error("write (2)");

	char *fic = mmap(NULL, i * sizeof(struct personne) + 8,// 8: alignement.
			  PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (fic == MAP_FAILED) error("mmap (write)");
	struct personne* fich = (struct personne*) (fic + 8);

	for (int j = 0; j <= i; j++)  /* criture des entres. */
		fich[j] = tab[j];

	err = munmap(fic, i * sizeof(struct personne) + 8);
	if (err) error("munmap (1)");
	fic = NULL; fich = NULL;

	err = close(fd);
	if (err) error("close (1)");

	/* Rcupration des donnes sauvegardes. */
	uint16_t nb_pers = 0; // Nombre de personnes dans l'annuaire.
	fd = open(ANNUAIRE, O_RDONLY);
	if (fd == -1) error("open (write)");

	err = read(fd, &nb_pers, 2);
	if (err != 2) error("read");

	fic = mmap(NULL, nb_pers * sizeof(struct personne) + 8, PROT_READ,
		   MAP_SHARED, fd, 0);
	fich = (struct personne*) (fic + 8);
	err = close(fd);
	if (err) error("close (2)");
	if (fich == MAP_FAILED) error("mmap (read)");

	for (int j = 0; j < nb_pers; j++) /* Affichage des entres. */
		printf("Entre %d\n\t %s %s (age: %d) -->  %s\n", j + 1,
		       fich[j].nom, fich[j].prenom, fich[j].age, fich[j].tel);

	err = munmap(fic, nb_pers * sizeof(struct personne) + 8);
	if (err) error("munmap (2)");
	return 0;
}
