#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define LINESIZE 200
#define MAXARGNUM 15

static inline int carspecial(char c) {
	if (c == ' ' || c == ';' || c == '&' || c == 0 || c == '\n' || c == '>')
		return 1;
	return 0;
}

static int separateurpresent(char *ligne, int *mustwait, int *fdout) {
	switch (ligne[0]) {
	case '&':
		*mustwait = 0;
		/* passe au travers */
	case ';':
		ligne[0] = '\0';
		return 1;
	case '>':
		ligne[0] = '\0';
		int i = 1;
		while (ligne[i] == ' ') {i++;}
		char* fichier = ligne + i;
		while (!carspecial(ligne[i])) {i++;}
		if (ligne[i] != 0)  ligne[i++] = 0;
		*fdout = open(fichier, O_WRONLY | O_CREAT | O_TRUNC, 0644);
		if (*fdout < 0) fprintf(stderr, "%s\n", strerror(errno));
		return i;
	}
	return 0;
}

static int comprendrecommande(char **argv, char *ligne, int offset, int *mustwait,
			      int *fdout)
{
	int j = 1;
	int i = offset;
	while (ligne[i] == ' ' || ligne[i] == ';' || ligne[i] == '&') i++;
	if (ligne[i] == 0) return -1;
	argv[0] = ligne + i;
	while (ligne[i] && ligne[i] != '\n'
	       && ligne[i] != ';' && ligne[i] != '&' && ligne[i] != '>') {
		if (ligne[i] == ' ') {
			ligne[i] = '\0';
			if (!carspecial(ligne[i+1])) {
				argv[j++] = ligne + i + 1;
				if (j == MAXARGNUM) return -10;
			}
		}
		i++;
	}
	argv[j] = NULL;
	int ret = separateurpresent(ligne + i, mustwait, fdout);
	if (ret) return i + ret;
	return 0;
}

static int executer(char* ligne) {
	char *argv[MAXARGNUM + 1];
	int offset = 0;
	do {
		int ret;
		int mustwait = 1;
		int fdout = -2;
		offset = comprendrecommande(argv, ligne, offset,
					    &mustwait, &fdout);
		if (offset == -1)   break;
		if (offset == -2)   continue;
		if (offset < -2) {
			fprintf(stderr, "Erreur de comprhension"
				" (trop d'arguments?) !");
			return -1;
		}

		ret = fork();
		if (ret < 0) {
			fprintf(stderr, "Impossible de forker !");
			exit(EXIT_FAILURE);
		}
		if (ret == 0) {
			if (fdout > 0) 
				dup2(fdout, STDOUT_FILENO);
			close(fdout);
			execvp(argv[0],argv);
			fprintf(stderr, "%s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
		if (fdout > 0) 
			close(fdout);
		if (mustwait)
			waitpid(ret, NULL, 0);
		else
			mustwait = 1;
		while (waitpid(-1, NULL, WNOHANG) > 0);
	} while (offset > 0);
	return 0;
}

static int lire(char* ligne) {
	int ret = read(STDIN_FILENO, ligne, LINESIZE);
	switch (ret) {
	case 0: /* entre ferme */
		exit(EXIT_SUCCESS);
	case LINESIZE: /* ligne longue */
		fprintf(stderr, "ligne trop longue: ignore\n");
		while (read(STDIN_FILENO, ligne, LINESIZE) >= LINESIZE);
		return -1;
	default: /* cas normal */
		ligne[ret - 1] = 0;
		break;
	}
	return 0;
}

static void afficheprompt(void) {
	printf(":");
	fflush(stdout);
}

int main()
{
	while(1) {
		char ligne[LINESIZE];
		afficheprompt();
		if (lire(ligne))	continue;
		if (executer(ligne))	continue;
	}
	return EXIT_FAILURE; /* non atteint */
}
