#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define _POSIX_C_SOURCE 2 /* Ncessaire sur certains Linux. */
#include <netdb.h> 

#define MAX_SIZE 256

void error(char * msg) { // cf. ligne /*@\ref{network::socket_TCP_client::error} du listing~\ref{network::socket_TCP_client}@*/
    perror(msg);
    exit(EXIT_FAILURE);
}
    /*** Gestion des interactions avec le serveur. ***/
void doit(int sockfd, struct addrinfo * serv) { 
    char      requete[MAX_SIZE] = "Luc, je suis ton pre!", reponse[MAX_SIZE];
    char      addr[INET6_ADDRSTRLEN], *ptr = "";  // Pour l'affichage de l'adresse.
    size_t    request_size = strlen(requete) + 1;
    in_port_t port; 

    ssize_t n = send(sockfd, requete, request_size, 0); // Envoi de la requte.
    if (n != request_size) error("[send]");
    switch (serv->ai_family) {
    case AF_INET:  {
	struct sockaddr_in  *sin  = (struct sockaddr_in *)  serv->ai_addr;
	ptr = (char *) inet_ntop(AF_INET, &sin->sin_addr, addr, sizeof(addr));
	port = sin->sin_port;
	break;     }
    case AF_INET6: {
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) serv->ai_addr;
	ptr = (char *) inet_ntop(AF_INET6, &sin6->sin6_addr, addr, sizeof(addr));
	port = sin6->sin6_port;
	break;     }
    }
    if (!ptr) printf("Requte envoye  %s port %d : <%s> (%ld octets envoys)\n",
		     ptr, port, requete, (long) n);

    n = recv(sockfd, reponse, MAX_SIZE-1, 0); // La rponse du serveur.
    if (n < 0) error("[recv]");
    reponse[n] = '\0'; // Prcaution pour l'affichage.
    printf("Rponse obtenue: <%s> (%lu octets)\n", reponse, (long) n);  
}

int main(int argc, char* argv[]) 
{
    char * servname = "localhost";    // Serveur par dfaut.
    char * port     = "6666";         // Port du service par dfaut.
    struct addrinfo hints, *res;
    int err, sockfd = -1;

    switch (argc) { // Traitement des arguments de la ligne de commande.
    case 3:  port     = argv[2];
    case 2:  servname = argv[1]; break;
    case 1:  break;
    default: fprintf(stderr, "Usage: %s [servname [port]]\n",argv[0]);
	     return EXIT_FAILURE;
    }
    /*** Construction de l'adresse de socket du serveur ***/
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_CANONNAME;  
    hints.ai_family = AF_UNSPEC;      
    hints.ai_socktype = SOCK_STREAM;  // Sockets connecte.
    if ((err = getaddrinfo(servname, port, &hints, &res)) < 0)  {
	if (err == EAI_SYSTEM) error("[getaddrinfo]");
	fprintf(stderr, "[getaddrinfo] on %s : %s\n", servname, gai_strerror(err));
	return EXIT_FAILURE;
    }
    for(; res != NULL; res = res->ai_next) {
	sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (sockfd < 0) continue;
	if (connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
	    close(sockfd);
	    sockfd = -1;
	    continue;
	}
	break; // Si on arrive ici, on a russi  se connecter au serveur.
    }
    if (sockfd  < 0)   error("[socket ou connect]"); 
    doit(sockfd, res); // Gestion des requtes.

    close(sockfd);     // Fermeture de la socket.
    freeaddrinfo(res); // Ne pas oublier de librer l'espace allou  res!
    return EXIT_SUCCESS;
}

