#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define MAX_SIZE 256
#define NUMPORT  6666     // Le port o les clients se connecteront.
#define TAILLE_FILE 14    // Nombre maximal de connexions dans la file d'attente.

void error(char * msg);      // Cf. ligne /*@\ref{network::socket_TCP_client::error} du listing~\ref{network::socket_TCP_client}@*/
void traitement(int sockfd); // Cf. listing /*@\ref{network::socket_TCP_server::doit}@*/

int main()
{
	int sock_ecoute, desc_conn;     // desc_conn: pour la socket de travail.
	struct sockaddr_in adr_svr;     // Pour l'adresse locale du serveur.
	struct sockaddr_in adr_client;  // Pour l'adresse du client qui se connecte.

        /*** Cration de la structure d'adresse locale du serveur. ***/
	memset(&adr_svr, 0, sizeof(adr_svr));
	adr_svr.sin_family      = AF_INET;
	adr_svr.sin_addr.s_addr = htonl(INADDR_ANY);
	adr_svr.sin_port        = htons(NUMPORT);

        /*** Cration de la socket serveur. ***/
	if ((sock_ecoute = socket(PF_INET,SOCK_STREAM,0)) == -1)
		error("[socket]");

	{/* On utilise la fonction setsockopt() pour autoriser la rutilisation de
	    l'adresse locale au redmarrage du serveur sans dlai. */
		int on = 1;
		if(setsockopt(sock_ecoute, SOL_SOCKET, SO_REUSEADDR, /*@\label{network::socket_TCP_server::setsockopt}@*/
			      &on, sizeof(on)))
			error("[setsockopt(SO_REUSEADDR)]");
	}

        /*** Attachement de la socket. ***/
	if (bind(sock_ecoute, (struct sockaddr *) &adr_svr, sizeof(adr_svr)))
		error("[bind]");

        /*** On en fait une socket d'coute. ***/
	if (listen(sock_ecoute, TAILLE_FILE)) /*@\label{network::socket_TCP_server::listen}@*/
		error("[listen]");
	printf("Serveur:  l'coute  l'adresse %s sur le port %d\n",
	       inet_ntoa(adr_svr.sin_addr), ntohs(adr_svr.sin_port));

	while(1) { // Boucle principale sur accept().
		socklen_t addr_len = sizeof(adr_client);
		desc_conn = accept(sock_ecoute, (struct sockaddr *) &adr_client, /*@\label{network::socket_TCP_server::accept}@*/
				   &addr_len);
		if (desc_conn == -1) {
			perror("[accept]");
			continue;
		}
		printf("Serveur: connexion accepte avec %s:%d\n", 
		       inet_ntoa(adr_client.sin_addr), ntohs(adr_client.sin_port));
		/*** Gestion des connexions acceptes avec des sous-processus. ***/
		int pid = fork();
		if (!pid) {         // Cration d'un processus fils. /*@\label{network::socket_TCP_server::fork}@*/
			close(sock_ecoute); /* Fermeture de la socket
					       d'coute par le fils. */
			traitement(desc_conn); // Traitement de la requte.
			return EXIT_SUCCESS;   // Sortie du processus fils.
		}
		close(desc_conn);          // Le pre ferme la socket de travail.
		if (pid == -1) { // Fin du programme si fork() choue.
			close(sock_ecoute);
			return EXIT_FAILURE;
		}
		// Nettoyage des processus fils.
		while(waitpid( -1, NULL, WNOHANG) > 0); /*@\label{network::socket_TCP_server::waitpid}@*/
	}
	close(sock_ecoute);   // Fermeture de la socket (jamais atteint...).
	return EXIT_SUCCESS;
}

void error(char * msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}

/** Traite une connexion accepte par envoi/reception de donnes */
void traitement(int sockfd)
{
	char requete[MAX_SIZE+1], reponse[MAX_SIZE] = "Nonnnnnnnnn !";
	size_t  answer_size = strlen(reponse)+1;
	size_t reste = answer_size;

	ssize_t n = read(sockfd, requete, MAX_SIZE-1); // Rception.
	if (n < 0) error("[read]");
	requete[n] = '\0'; // Prcaution pour l'affichage.
	/* On peut n'avoir reu qu'une requte partielle, pour certains programmes
	   il peut falloir faire d'autres read pour la complter. */
	printf(" - Requte reue: <%s> (%ld octets reus)\n", requete, (long) n);

	while (reste > 0) { // mission de la rponse
		n = write(sockfd, reponse + (answer_size - reste), reste);
		if (n < 0) error("[write]");
		printf(" - Rponse mise: <%s> (%ld octets envoys)\n",
		       reponse, (long) n);
		reste -= n;
	}
}
