#include <stdio.h> 
#include <stdlib.h> 
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <string.h>    
#include <arpa/inet.h>
#include <stdbool.h>
#include <sys/ioctl.h>
#include <errno.h>

#define MAX_SIZE 256
#define NUMPORT  6666    // Le port o les utilisateurs se connecteront 
#define BACKLOG  14      // Nombre maxi de connexions acceptes en file d'attente

void error(char * msg);      // cf. ligne /*@\ref{network::socket_TCP_client::error} du listing~\ref{network::socket_TCP_client}@*/

// ...
#undef  max
#define max(x,y) ((x) > (y) ? (x) : (y)) // Calcul du maximum
        /* Certains systmes ne fournissent pas la macro FD_COPY */
#ifndef FD_COPY
#  define FD_COPY(p, q)   memcpy((q), (p), sizeof(fd_set))
#endif

extern int errno;
void traitement(int sockfd, bool * close_conn) {
    char requete[MAX_SIZE], reponse[MAX_SIZE] = "Nonnnnnnnnn !";
    size_t  taille_rep = strlen(reponse)+1;    

    ssize_t n = recv(sockfd, requete, MAX_SIZE-1, 0); // Rception
    if (n < 0) {
	if (errno != EWOULDBLOCK) { perror("[recv]"); *close_conn = true; }
	return;
    }
    if (n == 0) { *close_conn = true; return; }
    requete[MAX_SIZE-1] = '\0'; // Prcaution pour l'affichage
    printf(" - Requte reue: <%s> (%ld octets reus)\n", requete, (long) n);

    size_t taille_reste = taille_rep;
    while (taille_reste > 0) {
	    n = send(sockfd, reponse, taille_reste, 0); // mission d'un bout de la rponse.
	    if (n == -1) {
		    perror("[send]");
		    *close_conn = true;
		    return;
	    } else
		    taille_reste -= n;
    }
}
int main() 
{
    int listenfd, connfd, i, max_fd = 0;
    struct sockaddr_in serv_addr, cl_addr; // IP du serveur et du client
    socklen_t len = sizeof(struct sockaddr_in);
    struct timeval timeout = { 1*60, 0 };  // timeout de 1 minute
    bool end_server = false;               // sentinelle de fin 
    
    memset(&serv_addr, 0, sizeof(serv_addr));  // initialisation @IP du serveur
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port        = htons(NUMPORT);

        /*** Cration de la socket, attachement et mise en coute ***/ 
    if ((listenfd = socket(PF_INET,SOCK_STREAM,0)) < 0) error("[socket]");
    const int on = 1;
    if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) 
	error("[setsockopt(SO_REUSEADDR)]");
    // On en fait une socket non bloquante; 
    if (ioctl(listenfd, FIONBIO, &on) < 0)                       error("[ioctl]"); /*@\label{network::socket_TCP_server_nonblockingIO::ioctl}@*/
    if (bind(listenfd, (struct sockaddr *) &serv_addr, len) < 0) error("[bind]");
    if (listen(listenfd, BACKLOG) < 0)                           error("[listen]");     
    printf("Serveur: en coute  l'adresse %s sur le port %d\n",
	   inet_ntoa(serv_addr.sin_addr), ntohs(serv_addr.sin_port));
        /*** Initialisation des ensembles de descripteurs ***/
    fd_set master_set, working_set; 
    FD_ZERO(&master_set);    FD_ZERO(&working_set);
    FD_SET(listenfd, &master_set);
    max_fd = max(max_fd, listenfd);

    do { 
	FD_COPY(&master_set, &working_set); // recopie de l'ensemble master_set
	    /*** Appel  select avec une attente jusqu'au timeout  ***/
	int r = select(max_fd+1, &working_set, NULL, NULL, &timeout);
	if (r < 0)  error("[select]");
	if (r == 0) { // timeout atteint
	    fprintf(stderr, "Timeout sur select() atteint: fin du programme\n");
	    break;
	}
	// Ici, r > 0 descripteurs sont prt. Il faut dterminer lesquels
	for (i=0; (i <= max_fd) && (r > 0); i++) {
	    if (FD_ISSET(i, &working_set)) {
		r--;  // on a trouv un descripteur prt en lecture
		if (i == listenfd) {
		    printf("La socket d'coute %d est prte en lecture...\n",i);
		    do { /*** On accepte tout connexion entrante ***/
			connfd = accept(listenfd,(struct sockaddr *)&cl_addr, &len);
			if (connfd < 0) {
			    if (errno != EWOULDBLOCK) { // accept a vraiment rat
				perror("[accept]");
				end_server = true;
			    }
			    break;
			}
			printf("Connexion %d accepte avec %s:%d\n", connfd, 
			       inet_ntoa(cl_addr.sin_addr), ntohs(cl_addr.sin_port));
			FD_SET(connfd, &master_set); 
			max_fd = max(max_fd, connfd);
		    } while (connfd != -1); 
		} else { // i est donc le descripteur d'une connexion entrante
		    bool close_conn = false; 
		    traitement(i, &close_conn); // Traitement de la requte
		    if (close_conn == true) {   // Il faut fermer proprement 
			close(i);
			FD_CLR(i, &master_set);
			if (i == max_fd)  // Mise  jour de max_fd
			    while(!FD_ISSET(max_fd, &master_set)) max_fd--;
		    }
		} // Fin connexion entrante active
	    } // Fin if (FD_ISSET(i, &working_set))
	} // Fin de la boucle for 
    } while (end_server == false);

    for (i=0; i < max_fd; i++) // Fermeture des sockets ouvertes
	if (FD_ISSET(i, &master_set)) close(i);
    return EXIT_SUCCESS;
}

void error(char * msg) {
    perror(msg);
    exit(EXIT_FAILURE);
}
