#include <stdio.h> 
#include <stdlib.h> 
#include <unistd.h>
#define __USE_MISC 1
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>    
#include <arpa/inet.h>

#define TAILLE_MAX 256
#define TAILLE_FILE 14       // Nombre maximal de connections dans la file d'attente.

void error(char * msg);      // Cf. ligne /*@\ref{network::socket_TCP_client::error} du listing~\ref{network::socket_TCP_client}@*/.
void traitement(int desc_sock); // Cf listing /*@\ref{network::socket_TCP_server::doit}@*/.
// [...]
#define SERVER_PATH "/tmp/server.socket"  // Chemin de la socket du serveur.

int main()
{
	int desc_ecoute, desc_conn;
	struct sockaddr_un adr_svr, adr_client;
	socklen_t lon = sizeof(struct sockaddr_un); 

	/*** Cration de l'adresse de socket Unix du serveur. ***/
	memset(&adr_svr, 0, sizeof(adr_svr));     // Remplissage  0.
#ifndef __GLIBC__   /* La glibc ne dfinit pas le champ sun_len */
	adr_svr.sun_len    = sizeof(adr_svr);
#endif
	adr_svr.sun_family = AF_LOCAL;
	strncpy(adr_svr.sun_path, SERVER_PATH, sizeof adr_svr.sun_path - 1); /*@\label{network::socket_Unix1_server::strncpy}@*/
	adr_svr.sun_path[sizeof adr_svr.sun_path  - 1] = 0;
    
	/*** Cration d'une socket ***/ 
	if ((desc_ecoute = socket(AF_LOCAL,SOCK_STREAM,0)) == -1)
		error("[socket]");

	/*** Attachement de la socket - utilisation de SUN_LEN ***/
	if (bind(desc_ecoute, (struct sockaddr *) &adr_svr, SUN_LEN(&adr_svr)))
		error("[bind]");

	/*** On en fait une socket d'coute. ***/
	if (listen(desc_ecoute, TAILLE_FILE))
		error("[listen]");
	printf("Serveur:  l'coute sur %s\n", adr_svr.sun_path);

	/*** Boucle principale pour accept(). ***/
	while(1) {
		memset(&adr_client, 0, sizeof(adr_client)); // Remplissage  0.
		desc_conn = accept(desc_ecoute, (struct sockaddr *) &adr_client,
				   &lon);
		if (desc_conn == -1) {
			perror("[accept]");
			continue;
		}
		/*** Gestion concurrente des connexions acceptes ***/
		int pid = fork();
		if (pid == -1) {
			perror("[fork]");
			close(desc_conn);
			continue;
		}
		if (!pid) {         // Cration d'un processus fils.
			close(desc_ecoute);    /* Fermeture de la socket
						  d'coute par le fils. */
			traitement(desc_conn); // Traitement de la requte.
			exit(EXIT_SUCCESS); // Sortie du processus fils.
		}
		close(desc_conn);          // Le pre ferme la socket de travail.
		while( waitpid( -1, NULL, WNOHANG) > 0);
	}
	close(desc_ecoute);   // Fermeture de la socket.
	return EXIT_SUCCESS;
}

void traitement(int desc_sock) {
	char requete[TAILLE_MAX+1], reponse[TAILLE_MAX] = "Nonnnnnnnnn !";
	size_t  answer_size = strlen(reponse)+1;

	ssize_t n = recv(desc_sock, requete, TAILLE_MAX-1, 0); // Rception.
	if (n < 0) error("[recv]");
	requete[TAILLE_MAX] = '\0'; // Prcaution pour l'affichage.
	printf(" - Requte reue: <%s> (%ld octets reus)\n", requete, (long) n);

	n = send(desc_sock, reponse, answer_size, 0); // mission de la rponse.
	if (n != answer_size ) error("[send]");
	printf(" - Rponse mise: <%s> (%ld octets envoys)\n", reponse, (long) n);
}



void error(char * msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}
